/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include<set>
#include<vector>

using namespace std;

#include "state.h"
#include "plan.h"
#include "task.h"
#include "search_entities.h"
#include "solver_context.h"
#include "search_context.h"
#include "fringe.h"
#include "utils.h"

/*
 *		Finds solution for a path or a edge in a DTG;
 */

static Task&task = Task::getInstance();

/**************************************************************************************/
/*********************************  LOCAL Helper actions  *****************************/

void build_FP(TransitionGraph& graph, unsigned* path, unsigned len){
	unsigned i,j;

	for(i=0;i<graph.size;i++)
		for(j=0;j<graph.size;j++)
			graph.edge[i][j].num_FP = 0;
		
	if( len <= 2 ) return;		//Only one edge, not necessary;
	
	for(i=len-2;i>0;i--){

		TransitionEdge& edge = graph.edge[path[i]][path[i+1]];
		TransitionEdge& preEdge = graph.edge[path[i-1]][path[i]];
		unsigned checkingFact = graph.vertex[path[i]][0];
		
		for(j=0;j<edge.num_SP;j++){
			assert(edge.SP[j]>=0 && edge.SP[j]< task.gnum_ft );
			if( task.fact_ordering[edge.SP[j]][checkingFact] 
			                                      && edge.SP[j]!=checkingFact ) //What's wrong here!!!?
				preEdge.FP[preEdge.num_FP++] = edge.SP[j];
		}
		
		for(j=0;j<edge.num_FP;j++)
			if( task.fact_ordering[edge.FP[j]][checkingFact] 
			                                      && edge.FP[j]!=checkingFact){
				preEdge.FP[preEdge.num_FP++] = edge.FP[j];
				if( preEdge.num_FP > MAX_FP_NUM ){
					cout<<"ERROR: Forced precondition array for Edge overflow, ";
					cout<<"Please increase MAX_FP_NUM"<<endl;
					exit(0);
				}
			}
	}
}

void sort_path_by_length(unsigned** path, unsigned* path_len, unsigned num){
	unsigned i,j;
	
	if( num < 2 ) return;//un-necessary;
	
	for( i = num-1 ; i > 0 ;i-- )
   	for(j = 0; j < i; j++){
   		if( path_len[j] > path_len[j+1]){
   			unsigned *tmpPath = path[j];
   			path[j] = path[j+1];
   			path[j+1] = tmpPath;

   			swap(path_len[j], path_len[j+1]);
   		}
   	}
}

/*********************************  LOCAL Helper actions  *****************************/
/**************************************************************************************/



void DTGSearch::set_dest( unsigned factId ){
	assert( task.gft[factId].graph == this->graph.id );
	
	this->toFact = factId;
	this->toIndex = task.gft[factId].vertex;
	hash_point = 0;
}

void DTGSearch::set_dest_index(unsigned index ){
	this->toIndex = index;
}

Plan DTGSearch::solve_path_no_merge(Plan& plan,
									unsigned **path,
									unsigned *path_len,
									unsigned path_num){
	unsigned i,j;
	unsigned fromIndex = toIndex = UNDEFINED;
	Plan result;
	result.set_infinity();

	for( i=0 ; i<path_num ; i++){
		if( context.isDebug() ) 
			printf("[%d]Executing Path %d NO_MERGE:\n",context.depth, i);

		result.set_empty();
		result.es.clone( plan.es );
		build_FP( graph, path[i], path_len[i] );
		//unsigned path_hash = DJBHash( path[i], path_len[i]);
		
		for(j=0;j<path_len[i]-1;j++){
			fromIndex = path[i][j];
			toIndex   = path[i][j+1];

			TransitionEdge& edge = graph.edge[fromIndex][toIndex];
			context.info("Running an edge.\n");
			
			context.depth++;
			Plan edgePlan = solve_edge( edge, result.es ); 
			assert( edgePlan.es.gvalue != result.es.gvalue );
			context.depth--;
			
			if( !edgePlan.is_infinity() ){
				result.merge( edgePlan );
				result.es = edgePlan.es;
			}else{
				result.set_infinity();
				break;
			}
		}
		
		if( !result.is_infinity() && (j == path_len[i]-1) ){
			if( context.isDebug() )
				printf("[%d]Breaking since valid dtgSoln found [%d]\n",
						context.depth, result.length() );
			break;
		}
	}
	
	return result;
}

/*
 *	plan.es ----- ? -----> this->toIndex
 */
Plan DTGSearch::solve(Plan& plan){
	
	if ( context.isDebug() ) printf("[%d]DTGSearch::solve()\n", context.depth);

	unsigned i, j, fromIndex=UNDEFINED;
	Plan result;
	fromIndex = plan.es.gvalue[ graph.id ];
	
	assert( plan.es.is_valid() );
	assert( fromIndex !=UNDEFINED && fromIndex < graph.size );
	assert( toIndex != UNDEFINED && toIndex < graph.size );
	
	if( fromIndex == toIndex  ){
		if( context.depth >0 ){
			cout<<"Something must be wrong"<<endl;
			exit(0);
		}
		result.set_empty();
		result.es.clone( plan.es );
		return result;
	}
	
	context.info( "DTGS:Preprocessing#1\n");
	//This operation needs one more condition;
	if( !graph.vertex[fromIndex].isNone() && 
			context.is_violating_ff( graph.vertex[fromIndex][0] ) ){
		result.set_infinity();
		return result;
	}
	
	context.info("DTGS:Preprocessing#2\n");
	if( graph.path != NULL ){
		char bb[250];
		context.info("Returning from DTG::solve() with inner-path\n");
		result.set_empty();
		result.es.clone( plan.es );
		Plan& addPlan = graph.path[fromIndex][toIndex];
		if( addPlan.is_infinity() ) return Plan::create_infinity_plan();

		result.merge( addPlan );
		for( i=0; i< addPlan.length() ; i++)
			result.es.apply_action_effect( addPlan[i] );

		sprintf(bb,"With plan length%d\n",result.length());
		context.info(bb);
		return result;
	}
	
	if( context.info("CP_DTG***") ) context.print_ff_list();
	
	unsigned** path = new unsigned*[MAX_DTG_SIZE];
	for( i=0;i<MAX_DTG_SIZE;i++ ) path[i] = NULL;
	unsigned* path_len = new unsigned[MAX_DTG_SIZE];
	unsigned num = graph.retrieve_all_path( fromIndex, toIndex, path, path_len );
	
	if( context.isDebug()  ){
		printf("[%d]Before sorting:%u parthes",context.depth, num);
		printf("[from%u to %u] in graph %u\n",fromIndex,toIndex,graph.id);
	}

	sort_path_by_length( path, path_len, num);
	assert( num < MAX_DTG_SIZE );

	if( context.isDebug() ){
		printf( "[%d]There are [%d] pathes found in total.\n", context.depth , num );
		for( i=0 ; i<num ; i++ ){
			printf( "[%d]PATHS:", context.depth );
			for( j=0;j<path_len[i];j++ )
				 graph.print_vertex( path[i][j]),printf(";");
			printf("\n");
		}
	}
	
	result = solve_path_no_merge( plan, path, path_len, num );
	
	for(i=0;i<num;i++) 
		delete[] path[i];
	delete[] path;
	delete[] path_len;

	return result;
}


void DTGSearch::process_action_list(	TransitionEdge& edge,
										State& currentState, 
									 	vector<unsigned>& actionArray ){
	
	bool deleteFlag[MAX_ACTION_IN_EDGE];
	unsigned i,j,action;
	for( i=0;i<edge.num_action;i++ ) 
		deleteFlag[i]=0;

	for(i=0;i<edge.num_action;i++)
		if( deleteFlag[i] == 0 ) 
			actionArray.push_back(  edge.action[i] );

	int sa[MAX_ACTION_IN_EDGE];
	memset( sa, 0 , actionArray.size() * sizeof(int) );
	
	//This heuristic counts vertex local distances
	for(i=0;i<actionArray.size();i++){
		action = actionArray[i];
		for(j=0;j<task.gop[action].num_T;j++){
			unsigned v_curr = currentState.gvalue[task.gop[action].Tg[j]];
			unsigned v_exp = task.gop[action].Tp[j]; 
			if( v_exp == NO_PRE_NEEDED ) continue;
			if( v_curr == v_exp ) continue;
			
			assert( task.transGraph[task.gop[action].Tg[j]].distance != NULL );
			sa[i] += task.transGraph[task.gop[action].Tg[j]].distance[v_curr][v_exp];			//What is this heuristic for? Storage?
			//if( v_curr == v_exp ) sa[i]++;
		}
	}
	
	for( i = actionArray.size()-1 ; i > 0 ;i-- )
   	for( j = 0; j < i; j++ ){

		if( sa[j] > sa[j+1] ){
			swap( sa[j], sa[j+1]);
			swap( actionArray[j], actionArray[j+1]);
		}   		
   	}
}



Plan DTGSearch::solve_edge(	TransitionEdge& edge, 
							State& passed_in_state){
	
	if( context.isDebug() ) 
		printf( "[%d]DTGS::edge()[graph%d]\n" , context.depth, graph.id );
	
	unsigned i,j,ft,action;
	Plan result;
	unsigned saved_FFN = context.num_fixed_fact;
	vector<unsigned> actionArray;
	unsigned preFact[MAX_FACTSET_SIZE];
	unsigned preFactSize = 0;
	vector<unsigned> info_vec;

	//To detour visited states;
	unsigned hv = passed_in_state.hash_value();
	hv = hv + this->hash_point;	
	set<unsigned>::iterator iter = context.cg_index.find( hv );	
	if( iter != context.cg_index.end() ) return Plan::create_infinity_plan();
	context.cg_index.insert(hv); 
	
	process_action_list(  edge, passed_in_state , actionArray );
	
	if( context.isDebug() ) {
		printf("[%d]Geting_Edge_Weight:", context.depth );printf(" FROM ");
		graph.print_vertex( edge.fromIndex);
		printf("to "); graph.print_vertex( edge.toIndex);
		printf("totally %d Actions\n",actionArray.size());
		for(i=0;i<actionArray.size();i++)
			task.print_op_name( actionArray[i]);
		printf("\n");
	}

	result.es.clone( passed_in_state );
	Plan cddt_plan;
	for( i = 0 ; i<actionArray.size() ; i++){
		action = actionArray[i];
		
		if ( context.isDebug() ){
			printf("[%d]CHECKING ACTION[%d,%d]", context.depth, i, action ); 
			task.print_op_name( action );
			printf("\n");
		}

		//Preprocessing #1: Del-Effect Violating fixed_list?
		for(j=0;j<task.gop[action].num_D;j++ )
			if( context.is_violating_ff( task.gop[action].D[j]) )
				goto NEXT_ACTION;
		
		// Pre-processing #3: Computing Forced-PRE-Conditions;
		preFactSize = 0;
		for(j=0;j<task.gop[action].num_P;j++)
			preFact[preFactSize++] = task.gop[action].P[j];

		for(j=0;j<edge.num_FP;j++){
			ft = edge.FP[j];
			if(  passed_in_state.gvalue[task.gft[ft].graph] == task.gft[ft].vertex) 
				preFact[preFactSize++] = edge.FP[j] ;
		}
		
		cddt_plan.set_empty();
		if ( !result.es.isSatisfied( preFact, preFactSize ) ){
			context.depth++;
			FactsetSearch fss( preFact, preFactSize, context, action );
			vector<unsigned> info_vec;
			info_vec.push_back( graph.id );
			info_vec.push_back( action );			//A new hash point;
			fss.hash_point = DJBHash( info_vec );		
			cddt_plan = fss.solve( result.es );
			context.depth--;
		}

		//	Post-processing the result;		
		if( !cddt_plan.is_infinity()   
			&& !context.close_list.exists(cddt_plan.es) 
			&& !context.fringe_ptr->constains(cddt_plan.es)  ){
			
			result.merge( cddt_plan );
			if( !cddt_plan.is_empty() ){
				result.es = cddt_plan.es;
			}
			
			assert( result.es.gvalue != NULL && result.es.gvalue[0]!=UNDEFINED );
			result.add_oper( action );
			result.es.apply_action_effect( action );
			return result;
		}

NEXT_ACTION:
		context.num_fixed_fact = saved_FFN;
		result.es.clone( passed_in_state );
	}

	context.info( "Edge fail, returning INF.\n" );
	return Plan::create_infinity_plan();
}

//END-OF-FILE;
