
/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include "state.h"
#include "plan.h"
#include "task.h"
#include "search_entities.h"
#include "solver_context.h"
#include "search_context.h"
#include "fringe.h"
#include "utils.h"
#include "search_helper.h"

using namespace std;


static Task& task = Task::getInstance();
void sort_facts( unsigned *graphs, unsigned *vertex, unsigned size);



FactsetSearch::FactsetSearch(unsigned* fs, unsigned s, SearchContext& sc, unsigned fa):
	context(sc),fromAction(fa),size(s){
	for (unsigned i=0; i<size; i++)
		factset[i] = fs[i];
	hash_point = 0;
}

/*
 * 	If find out a valid plan then return with a Plan, which has .$seq and .$es defined.
 */
Plan FactsetSearch::solve_fixed_order(	State& passed_in_state, 
										unsigned* gIndex, unsigned *vIndex	) {

	if (context.isDebug() )
		printf("[%d]FS:fixed_order, with %d facts\n", context.depth, size );

	Plan result;
	unsigned saved_FFN = context.num_fixed_fact;
	unsigned executing = 0, firstUnsat = size, fact = UNDEFINED;

	for(firstUnsat=0; firstUnsat<size; firstUnsat++)
		if (passed_in_state.gvalue[gIndex[firstUnsat]]!= vIndex[firstUnsat])
			break;

	
	/* This iteration terminates only if:
	 * 			the there cannot be any solution for the first desired/UNSAT fact 
	 * It could fall into infinity loop, however, if it does happen, 
	 * */
	
	do {
		result.set_empty();
		result.es.clone( passed_in_state );
		context.num_fixed_fact = saved_FFN;

		context.info("Before a new iteration.\n");
		for ( executing=0; executing<size; executing++ ) {

			unsigned exeG = gIndex[executing];
			unsigned exeV = vIndex[executing];

			TransitionGraph& graph = task.transGraph[exeG];
			fact = graph.vertex[exeV][0];
			assert( task.gft[fact].graph  == exeG );
			assert( task.gft[fact].vertex == exeV );
			
			if ( exeV == result.es.gvalue[exeG]){
				context.add_ff(fact);
				continue;
			}
			
			if ( context.recurMark[graph.id] ){
				result.set_infinity();
				break;
			}

			if ( context.isDebug() ) {
				printf("[%d]cffo:EXECUTING fact{%d/%d}", context.depth, executing, size);
				task.print_ft_name(fact),printf("\n");
			}
			
			context.depth++;
			context.recurMark[graph.id] = true;
			DTGSearch ds( graph, context );
			ds.hash_point = this->hash_point;
			ds.set_dest_index( exeV );
			Plan partialPlan = ds.solve( result );

			context.recurMark[graph.id] = false;
			context.depth--;
			
			if( partialPlan.is_infinity() ){
				context.info("partialPlan is infinity.\n");
				break;
			}
			
			result.merge( partialPlan );
			result.es = partialPlan.es;
			
			if ( context.isDebug() ){
				printf("[%d]FACTSET-Merging plan:", context.depth);
				result.println();
			}
			context.add_ff( graph.vertex[exeV][0] );
		}
		
		if ( executing == firstUnsat || executing < size ){
			result.set_infinity();
			break;
		}
	}while( result.is_infinity() );
	
	//If it's naively satisfied then cgplan_factset_atomic must have detected and returned;
	assert( !result.is_empty() );
	
	context.info("FS::fixed_order() returning.\n");
	return result;
}



Plan FactsetSearch::solve(State& passed_in_state){

	context.info("FUNCTION:cgplan_factset::solve()\n");

	unsigned i,j;
	unsigned orderedGraph[MAX_FACTSET_SIZE],orderedVertex[MAX_FACTSET_SIZE];
	unsigned graphIndex[MAX_FACTSET_SIZE],vertexIndex[MAX_FACTSET_SIZE];
	unsigned saved_FFN = context.num_fixed_fact;
	//unsigned perm_count = 0;
	Permutation perm;
	Plan result;

	unsigned sas_size = translate_facts( graphIndex, vertexIndex );	
	assert( sas_size == size );   // inequal in those multi-dimensional graphs;

	if( context.isDebug() ){
		printf("[%d]List of facts(goal/precond) to be satisfied:", context.depth);
		print_factset( factset, size );
	}

	sort_facts( graphIndex, vertexIndex, sas_size );
	perm.init_perm_status( sas_size );
	//vector<unsigned> history;
	//unsigned ident;
	
	//Here we have a simple heuristic to break out the permutation loop (using magic number);
	//However, when the precondition increases, this estimation of constant number may 
	//cause problems, so I change it to compare with $sas_size$;
	unsigned feasible_ordering = 0;
	while( !perm.done ){
		
		context.info( ">>Doing A new permutated setting.\n" );
		
		for(i=0;i<size;i++){
			orderedGraph[i]  = graphIndex[perm.a[i]];
			orderedVertex[i] = vertexIndex[perm.a[i]];
		}
		
		if( context.isDebug() )
			cout<<perm.a[0]<<","<<perm.a[1]<<endl;
		
		//filter#1:
		/*	Comment: This part seems to be less helpful in Openstack */
		for(i=0;i<size;i++ )
		for(j=i+1;j<size;j++){
			unsigned fact1 = task.transGraph[orderedGraph[i]].vertex[orderedVertex[i]][0];
			unsigned fact2 = task.transGraph[orderedGraph[j]].vertex[orderedVertex[j]][0];
			if( fact1 >= task.gnum_ft || fact2 >= task.gnum_ft ) continue;
			if( task.fact_ordering[fact2][fact1] )
				goto next_permutation;
		}
		
		//filter#2: It turns out to be not inaccuracy in some domains;
		/*
		ident = 0;
		for(i=0;i<size;i++)
			ident = ident << 3 + orderedGraph[i];
		for(i=0;i<history.size();i++)
			if( history[i] == ident ) break;
		if( i <= history.size() ) 
			goto next_permutation; */

		if( context.isDebug()){
			printf("[%d]Ordered factset:",context.depth);
			for(i=0;i<size;i++){
				task.print_ft_name( orderedGraph[i], orderedVertex[i] );
				printf(";");
			}
			printf("\n");
		}
		
		feasible_ordering++;
		result = solve_fixed_order( passed_in_state, orderedGraph, orderedVertex  );
		
		if( !result.is_infinity() ){
			if( context.info("FS:solve returning.") )  result.println();
			
			context.num_fixed_fact = saved_FFN;
			assert( result.es.gvalue!= NULL );
			return result;
		}

		context.info("Before continue with next permutation\n");
		
next_permutation:
		if( perm.done && context.do_permutation ) break;
		//This threshold value actually is supposed to around 2 or 3 times of dlevel value; 
		//Now just hack it with a magic number;
		if( !context.do_permutation && ( feasible_ordering > sas_size|| context.depth>3  ) ) break;
		
		perm.make_next_permutation();
		context.num_fixed_fact = saved_FFN;
	}
	
	context.info("FactsetSearch::solve() returns infinity.\n");
	return Plan::create_infinity_plan();
}


/*
 * Local method, used only in this .cpp file;
 */
unsigned FactsetSearch::translate_facts( unsigned *graphs,  unsigned* vertice ){
	unsigned i;
	
	unsigned rtn = 0;
	for(i=0;i<size;i++){
		assert( factset[i]>=0 && factset[i]<= task.gnum_ft + task.transGraphNum );
		
		if( factset[i]<0 || factset[i]>= task.gnum_ft + task.transGraphNum ){
			printf("C_ERROR:%d %d, %d",i, factset[i], task.gnum_ft + task.transGraphNum );
			exit(0);
		}
		
		unsigned graphId;
		if( factset[i] < task.gnum_ft)
			graphId= task.gft[factset[i]].graph;
		else 
			graphId = factset[i] - task.gnum_ft;
		
		TransitionGraph& graph = task.transGraph[graphId];
		unsigned vid = UNDEFINED;
		if( graph.dimension == 1){
			if( factset[i] < task.gnum_ft )
				vid = task.gft[factset[i]].vertex;
			else if( factset[i] < task.gnum_ft + task.transGraphNum )
				vid = task.transGraph[graphId].size - 1;
			else
				exit(0);		//must be something wrong;
		}else{
			//to-do: For composited graphs;
			assert( false );
		}
		
		assert( vid>=0 );
		graphs[rtn] = graphId;
		vertice[rtn] = vid;
		rtn++;
	}

	return rtn;
}



void sort_facts( unsigned *graphs, unsigned *vertex, unsigned size){
	unsigned i,j;

	for( i = size-1 ; i > 0 ;i-- )
   	for(j = 0; j < i; j++){
		unsigned graph1 = graphs[j+1];
		unsigned graph2 = graphs[j];
		
		unsigned vertex1 = vertex[j+1];
		int vertex2 = vertex[j];
		assert( graph1!= UNDEFINED && graph2!= UNDEFINED );
		
		unsigned fact1 = task.transGraph[graph1].vertex[vertex1][0];
		unsigned fact2 = task.transGraph[graph2].vertex[vertex2][0];
		
		if( fact1 >= task.gnum_ft || fact2 >= task.gnum_ft )  continue;

		//if( forced_ordering[fact2][fact1] || casualDependencyMatrix[graph1][graph2] ){
		if( task.causalMatrix[graph1][graph2] || 
			task.transGraph[graph1].depend_degree() > task.transGraph[graph2].depend_degree()){
			
			int tmp = graphs[j];
			graphs[j] = graphs[j+1];
			graphs[j+1] = tmp;
			
			tmp = vertex[j];
			vertex[j] = vertex[j+1];
			vertex[j+1] = tmp;
		}
   	}
}


