/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include<iostream>
#include<vector>
#include<map>
#include<list>

#include "dp.h"
#include "state.h"
#include "plan.h"
#include "task.h"
#include "utils.h"
#include "search_entities.h"
#include "search_context.h"
#include "solver_context.h"
#include "plan.h"
#include "fringe.h"

using namespace std;

static Task&task = Task::getInstance();

unsigned MAX_FRINGE_SIZE = 5000;
unsigned MAX_FRINGE_SIZE_BOUND = 30000;


unsigned hit_count = 0;

//to-do: we need a up bound for the heuristic value database;
unsigned Fringe::compute_heuristic( FringeElem& elem){
	
	if( elem.heuristic != UNDEFINED ) 
		return elem.heuristic;

	map<unsigned,unsigned>::iterator iter = deferred_index.find( elem.hash );
	if( iter!= deferred_index.end() ) {
		//hit_count++;
		//cout<<"HIT"<<endl;
		return iter->second;
	}
	
	int v = get_ff_heuristic( elem.plan.es );
	if( v == -1 ) v = INF;
	
	elem.heuristic = (unsigned)v;
	return elem.heuristic;
}


bool Fringe::find_slot( unsigned cost, int& pos){
	if( slots.size() == 0 ){
		pos =0;
		return false;
	}
		
    int low = 0;
    int high = (int)slots.size()-1;
    while (low < high) {
        int mid = (low + high)/2;
        if ( slots[mid].cost < cost )
            low = mid + 1;
        else
             high = mid; 
    }
    
    pos = low;
    //Exception of pos after binary search;
    if( cost > slots[slots.size()-1].cost ) pos = slots.size();
    
    if ( low < (int)this->slots.size() &&  slots[low].cost == cost )
    	return true;
    else
    	return false;
}


bool Fringe::insert( FringeElem& node ){
	int pos;
	bool found;
	unsigned new_cost=UNDEFINED;
	if( count() > MAX_FRINGE_SIZE ) 
		shrink();
	node.hash = node.plan.es.hash_value();	
	
	map<unsigned, FRINGE_INDEX_TYPE>::iterator iter = index.find( node.hash );
	
	if( iter != index.end() ){

		FRINGE_INDEX_TYPE& index_unit = iter->second;
		int old_f = index_unit.first;
		int old_h = index_unit.second;
		//int old_g = old_f - old_h;
		new_cost = node.heuristic = old_h;
		
		if( old_h < (int)node.plan.cost() ) return false;
		
		found = find_slot( old_f , pos );
		assert( found );
		assert( (int)slots[pos].cost == old_f );
		assert( pos >=0 && pos<=(int)slots.size() );
		
		FringeSlot& slot = slots[pos];

		found = slot.find_bucket( old_h, pos, false );		
		assert( found );	
		assert(  pos >=0 && pos <= (int)slot.buckets.size() );
		
		FringeBucket& bucket = slot.buckets[pos];
		list<FringeElem>::iterator iter3 = bucket.elems.begin();
		for(;iter3!=bucket.elems.end();iter3++){					
			if( iter3->hash == node.hash ){
				bucket.elems.erase(iter3);
				slot.elem_count--;
				elem_count--;							//Don't need to take care of index here;
				break;
			}
		}
		
		new_cost += node.plan.cost();
		iter->second.first = new_cost;
		
	}else{

		new_cost = compute_heuristic ( node );
		assert( new_cost != UNDEFINED );
		if( new_cost == (unsigned)INF ){
			context.close_list.add( node.plan.es );
			return false;
		}
		
		new_cost += node.plan.cost();
		index[node.hash] = make_pair( new_cost, node.heuristic );				//Change to cost;
	}
	
	assert( new_cost < (unsigned)INF  );
	
	//If the state is a new one;
	found = find_slot( new_cost, pos );
	bool succ = false;
    if ( found )
    	succ = slots[pos].push( node );
    else{
    	FringeSlot slot;
    	slot.cost = new_cost;
    	succ = slot.push( node );
    	
    	slots.insert( slots.begin() + pos, slot );
    }
    
    if( succ )
    	elem_count++;
    else
    	deferred_index[node.hash] = node.heuristic;
    
//    check_count_consistency();
    return succ;
}


FringeElem Fringe::pop(){
	unsigned i,j;
	for(i=0;i<slots.size();i++){
		FringeSlot& slot = slots[i];
		if( slot.count() == 0 ) continue;
		
		for(j=0;j<slot.buckets.size();j++){
			FringeBucket& bucket = slot.buckets[j];
			
			if ( bucket.elems.size() !=0 ){
				FringeElem elem = bucket.elems.front();

				bucket.elems.pop_front();
				index.erase( elem.hash );
				elem_count--;
				slot.elem_count--;
				//check_count_consistency();
				return elem;
			}
		}
	}
	cout<<"ERR: Poping from a empty queue."<<endl;
	assert(false);
	exit(0);
}

//Cut the number of pending state in the fringe into half (in terms of fringe)
unsigned Fringe::shrink(){
	unsigned i,j;
	unsigned count = 0;
	
	bool first_found = false;
	for( i = 0 ; i<slots.size();i++){
		FringeSlot& slot = slots[i];
		if( !first_found && slot.count() != 0 ){
			first_found = true;
			continue;
		}
		
		unsigned shr_slot = 0; 
		for(j=0;j<slot.buckets.size();j++)
			shr_slot += slot.buckets[j].shrink();
		slot.elem_count -= shr_slot;
		count += shr_slot;
	}
	elem_count -= count;
	
	if( task.solverContext->debug )
		cout<<count<<"states removed"<<endl;
	
	if( MAX_FRINGE_SIZE < MAX_FRINGE_SIZE_BOUND) 
		MAX_FRINGE_SIZE += 1627;
	
	return count;
}



bool Fringe::is_empty(){
	
	for(unsigned i=0;i<slots.size();i++){
		if( slots[i].count() > 0 ) 
			return false;
	}
	
	return true;
}

bool Fringe::constains(State& state){
	map<unsigned, FRINGE_INDEX_TYPE>::iterator iter = index.find( state.hash_value() );
	if( iter != index.end() )
		return true;
	else
		return false;
}

unsigned Fringe::expand_all_succ(FringeElem& f_elem, unsigned scope){
	
	if( selected_graphs.size() == 0 ){
		cout<<"ERR:No graphs incorported in this search"<<endl;
		exit(0);
	}
	assert( selected_graphs.size() > 0 );
	
	unsigned count = 0,gid,i,j,k, val;
	State& state = f_elem.plan.es;
   	
	vector<TransitionEdge*> change_history;
	for( k = 0; k < selected_graphs.size(); k++ ){
		gid = selected_graphs[k];
		TransitionGraph& graph = task.transGraph[gid];
		val = state.gvalue[gid];

		for(i=0;i<graph.size;i++){
			TransitionEdge& edge = graph.edge[val][i];
			if( edge.num_action == 0 || i == val ) continue;
			
			for( j=0; j<change_history.size(); j++ )
				if( change_history[j] == &(edge) ) break;
			if( j < change_history.size() ) continue;
			
			DTGSearch d_search( graph, context );
			//cout<<"Working edge";edge.print();cout<<endl;
			//printf("\n***CG_Search Debug_Start***\n"); 		
			//context.print_level = 10;
			context.cg_index.clear();
			Plan partialSoln = d_search.solve_edge( edge, state );
			//context.print_level = 10;
			
			if( 	partialSoln.is_infinity() || 
					partialSoln.is_empty() || 
					context.close_list.exists( partialSoln.es ) 	)
				continue;
			
			assert( !partialSoln.is_infinity() );
			assert( !f_elem.plan.is_infinity() );
			
			Plan result;
			result.merge( f_elem.plan );
			result.merge( partialSoln );
			result.es = partialSoln.es;
			result.evaluate_cost();	
			
			for( j=0;j<task.transGraphNum;j++ )
				if( state[j]!=result.es[j] ){
					TransitionEdge* e_ptr = 
						&(task.transGraph[j].edge[state[j]][result.es[j]]);
					change_history.push_back( e_ptr );
				}
			
			assert( result.es[0] != UNDEFINED );
			FringeElem elem( result );
			elem.edge = &(graph.edge[val][i]);
			
			insert( elem );
			count++;
		}
	}

	return count;
}

bool Fringe::select_graphs( unsigned scope){
	/*		
	bool expd_flag[MAX_DTG_NUM];
	bool goal_atd[MAX_DTG_NUM];
	memset( goal_atd, false, sizeof(bool) * task.transGraphNum );
	memset( expd_flag, false, sizeof(bool) * task.transGraphNum );


	for(i=0;i<goals.size();i++){
		gid = task.gft[goals[i]].graph;
		if( f_elem.plan.es[gid] == task.gft[goals[i]].vertex )
			goal_atd[gid] = true;
	}
	for( i=0;i<task.transGraphNum;i++){
		if( !goal_atd[i] && task.transGraph[i].dlevel==0){
			expd_flag[i]=true;
			continue;
		}
		if( task.transGraph[i].dlevel <= scope ){
			for(j=0;j<task.transGraphNum;j++)
				if( task.transitCausalMatrix[j][i] && !goal_atd[j] ){
					expd_flag[i] = true;
					break;
				}
		}
		} */

	unsigned i,j,gid;
	
	assert( selected_graphs.size() == 0 );
	for(gid=0;gid<task.transGraphNum;gid++)
		if( task.transGraph[gid].dlevel <= scope ) 
			selected_graphs.push_back( gid );
	assert( selected_graphs.size() > 0 );
	sort( selected_graphs.begin(), selected_graphs.end());
	
	if( task.solverContext->display_info == 300){
		for(i=0;i<selected_graphs.size();i++)
			cout<<selected_graphs[i]<<";";
		cout<<endl;
	}
	
	//eliminate all the duplicated graphs regarding Accompany Relations;
	for( i=0; i<goals.size(); i++ ){
		gid = task.gft[ goals[i] ].graph;
		for( j=0; j<task.aco_pr_vec.size() ; j++ )
			if( task.aco_pr_vec[j].first == goals[i])
				break;
		if( j == task.aco_pr_vec.size() ) continue;		//not found;
		
		vector<unsigned>& slaves = task.aco_pr_vec[j].second;
		//If all the slaves are included in this scope then we can ignore goals[i];
		for(j=0;j<slaves.size();j++){
			unsigned s_gid = task.gft[slaves[j]].graph;
			//cout<<"checking slave";task.print_ft_name( slaves[j] );cout<<"gid"<<s_gid<<endl;			
			assert( s_gid != UNDEFINED );
			if ( bsearch_vector( selected_graphs, s_gid) == -1 ) 
				break;
			//cout<<" is in scope"<<endl;
		}
		
		if( j == slaves.size() ){	
			vector<unsigned>::iterator iter = 
				find( selected_graphs.begin(), selected_graphs.end(), gid );
			selected_graphs.erase( iter);
		}
	}
	
	assert( selected_graphs.size() > 0 );
	
	//Sort them by dlevel, lower level first;
	for( i = selected_graphs.size() - 1 ; i > 0 ; i-- )
   	for( j = 0 ; j < i; j++ )
	if( task.transGraph[selected_graphs[j]].dlevel < 
			task.transGraph[selected_graphs[j+1]].dlevel ){
		swap( selected_graphs[j], selected_graphs[j+1] );
	}
	
	
	if( task.solverContext->display_info == 300){
		cout<<"The graphs that in the scope of this search:"<<endl;
		for(i=0;i<selected_graphs.size();i++){
			cout<<selected_graphs[i]<<",";
		}
		cout<<endl;
		exit(0);
	}
	
	return true;
}


/*************************  affli routines **************************/

Fringe::Fringe(SearchContext& sc, vector<unsigned>& gvec):
			context(sc),goals(gvec){
	elem_count = 0;
	
	expand_count = 0;
	href_count = 0;
	expand_conflict = 0;	
}


Fringe::~Fringe(){
	
}

unsigned Fringe::count(){
	return elem_count;
}


void Fringe::print_status(){
	unsigned i,j;	
	cout<<"HIT COUNT"<<hit_count<<endl;
	cout<<"Number of slots:"<<slots.size()<<endl;
	for(i=0;i<slots.size();i++){
		FringeSlot& slot = slots[i];
		cout<<"SlotCost"<<slot.cost<<endl;
		for(j=0;j<slot.buckets.size();j++){
			list<FringeElem>::iterator iter = slot.buckets[j].elems.begin();
			for(;iter!=slot.buckets[j].elems.end();iter++){
				printf("[%u/%u:]", iter->plan.cost() + iter->heuristic , 
									iter->heuristic   );
				
				assert( iter->heuristic + iter->plan.cost() == slot.cost );
				assert( iter->heuristic == slot.buckets[j].heuristic );
			}
			cout<<";";
		}
		cout<<endl<<endl;
	}
}




void Fringe::print_statistics(){
	for(unsigned i=0;i<slots.size();i++){
		FringeSlot& slot = slots[i];
		printf( "Cost %u(#%u)",slot.cost, slot.count() );
	}
	cout<<endl;
}

void Fringe::check_count_consistency(){
	unsigned count = 0;
	for(unsigned i=0;i<slots.size();i++){
		unsigned count1 = 0;
		for(unsigned j=0;j<slots[i].buckets.size();j++)
			count1 += slots[i].buckets[j].elems.size();
		if(  count1 != slots[i].elem_count )
		cout<<"#"<<i<<"Supposed to be "<<count1<<"But is "<<slots[i].elem_count<<endl;
		assert( count1 == slots[i].elem_count);
		count += count1;
	}
	assert( count == elem_count );
}

//END of file;


