/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include "state.h"
#include "plan.h"
#include "task.h"
#include "search_entities.h"
#include "solver_context.h"
#include "search_context.h"
#include "fringe.h"
#include "utils.h"


using namespace std;


unsigned determine_graph_levels(vector<unsigned>& goals);
static Task&task = Task::getInstance();

GoalSearch::GoalSearch(vector<unsigned> g_vec){
	
	unsigned i,fact,j;
	for(i=0;i<g_vec.size();i++){
		orig_goals.push_back( g_vec[i] );
		goals.push_back( g_vec[i] );
	}
	
	soln_count = 0;
	best_soln.set_infinity();
	
	gt_acco = false;	
	skip_first_ss = false;		
	
	//to-do: transform goals with conditions;
	unsigned count = 0;
	for( i=0;i<task.orig_goals.size();i++){
		fact = task.orig_goals[i];
		for(j=0;j<task.aco_pr_vec.size();j++)
			if( fact == task.aco_pr_vec[j].first )
				break;
		if( j < task.aco_pr_vec.size())  count++;
	}
	if( count >= task.orig_goals.size()/2 )
		gt_acco = true;
	
	skip_first_ss = true;	
	for( i=0; i<task.orig_goals.size(); i++ ){
		fact = task.orig_goals[i];
		unsigned gid = task.gft[fact].graph;
		
		if( task.transGraph[gid].size != 2 )  //i.e. the interleave of goals is guaranteed;
			skip_first_ss = false;
	}
	
	if( gt_acco ) cout<<"[GT_ACCO]";
	if( skip_first_ss ) cout<<"[SKIP_SS]";
	cout<<endl;
}

Plan GoalSearch::solve(State& currentState){
	
	Plan result;
	unsigned max,scope,begin;
	SolverContext* sc = task.solverContext;
	
	if( gt_acco ){
		result = solve_atomic_accompany_goals( currentState );
		if( !result.is_infinity() ) update_solution( result );
	}else{
		result = solve_atomic_fixed_order( currentState );
		if( !result.is_infinity() ) update_solution( result );
	}

	if( sc->search_strategy == SS_HEURISTIC ){
		cout<<endl<<"**Using guided heuristic search(incremental)**"<<endl;
		
		max = determine_graph_levels( goals );
		
		if( sc->fixed_scope != UNDEFINED){
			cout<<endl<<"**Guided heuristic search (fixed scope "
				<<sc->fixed_scope<<")**"<<endl;
			result = solve_guided_search( currentState, sc->fixed_scope );
			return result;
		}
		
		begin = skip_first_ss ? 1 : 0;
		for( scope = begin; scope<= max; scope++ ){
			result = solve_guided_search( currentState, scope );
			if( !result.is_infinity() ){
				update_solution( result );
			}
		}
	}
	
	return this->best_soln;
}


/*
 * to-do:
 * 	Compile current goal into constraints and generate new goals;
 */
void GoalSearch::reduce_goals(){

}

/*
 * The goals may not be the original ones.
 */



Plan GoalSearch::solve_guided_search(State& state, unsigned scope){
	
	if( task.solverContext->display_info  == 108 ){
		for(unsigned i=0;i<task.transGraphNum;i++)
			if( task.transGraph[i].dlevel <= scope )
				task.transGraph[i].print();
		exit(0);
	}
	
	CloseList close_list;
	SearchContext context(close_list);
	context.do_permutation = true;
	context.cg_index.clear();
	Fringe fringe( context, this->goals );
	fringe.select_graphs( scope );
	context.fringe_ptr = &fringe;
	
	State initState;
	initState.clone( state );
	Plan empty_plan;
	empty_plan.set_empty();
	empty_plan.es = initState;

	FringeElem fringe_elem( empty_plan );	
	fringe.insert( fringe_elem );
	
	unsigned count = 0, dup_count = 0, expd_count = 0;
	while ( true ){
		
		if( fringe.is_empty() ) 
			return Plan::create_infinity_plan();
		FringeElem node = fringe.pop();
		
		if( node.plan.es.check_goal( goals) ) {
			if( task.solverContext->debug ) 
				cout<<endl<<"Totally "<<count<<" states visited;"<<dup_count<<"duplicates;"
				<<expd_count<<" expanded;"<<endl;
			cout<<"Abstract Search found a solution with cost "<<node.plan.cost()<<endl;
			return node.plan;
		}

		if ( !close_list.add( node.plan.es ) ){
			count++;
			expd_count += fringe.expand_all_succ( node, scope );
		}else
			dup_count++;
	}
}



/*
Plan GoalSearch::solve_one_accompany_goal(State& state, SearchContext& context){
	unsigned fact,sfact, i,j;
	Plan result;
	result.set_empty();
	result.es.clone( state );
	
	vector<unsigned> ffs;
	for(j=0;j<context.num_fixed_fact;j++) ffs.push_back( context.fixed_fact[j]);
	
	for(i=0;i<goals.size();i++){
		if( state.gvalue[task.gft[goals[i]].graph] == task.gft[goals[i]].vertex)
			continue;
		bool rtn = task.test_reachability( state, ffs, goals[i] );
		if ( !rtn ) 
			return Plan::create_infinity_plan(); 
	}
	
	for(i=0;i<goals.size();i++){
		fact = goals[i];
		if( state.gvalue[task.gft[fact].graph] == task.gft[fact].vertex)
			continue;
		
		if( task.solverContext->debug == true ){
			cout<<"Solving sub goal:";
			task.print_ft_name(fact);
			cout<<endl;
		}

		for(j=0;j<task.aco_pr_vec.size();j++)
			if( fact == task.aco_pr_vec[j].first ) break;
		if( j == task.aco_pr_vec.size()) continue;
		
		vector<unsigned>&slave_goals = task.aco_pr_vec[j].second;		
		for(j=0;j<slave_goals.size();j++){
			sfact = slave_goals[j];
			
			TransitionGraph& graph = task.transGraph[task.gft[sfact].graph];
			DTGSearch dtgsearch( graph, context );
			dtgsearch.set_dest( sfact );
			context.cg_index.clear();
			Plan soln = dtgsearch.solve( result );	
			
			if( soln.is_infinity() ) continue;
			
			context.add_ff( sfact );
			Plan other = solve_one_accompany_goal( soln.es, context );
			context.num_fixed_fact--;
			
			if( !other.is_infinity() && other.es.check_goal(goals) ) {
				result.set_empty();
				result.merge( soln );
				result.merge( other );
				result.es = other.es;
				return result;
			}
		}
		
		if( j == slave_goals.size() )
			return Plan::create_infinity_plan();
	}
	
	unsigned saved_FFN = context.num_fixed_fact;
	
	//Solve all the other goals;
	for(i=0;i<goals.size();i++){
		fact = goals[i];
		if( state.gvalue[task.gft[fact].graph] == task.gft[fact].vertex)
			continue;

		for(j=0;j<task.aco_pr_vec.size();j++)
			if( fact == task.aco_pr_vec[j].first ) break;
		if( j < task.aco_pr_vec.size()) continue;
		
		TransitionGraph& graph = task.transGraph[task.gft[fact].graph];
		DTGSearch dtgsearch( graph, context );
		context.cg_index.clear();
		dtgsearch.set_dest( fact );
		result = dtgsearch.solve( result );
	}
	
	context.num_fixed_fact = saved_FFN;
	if( result.es.check_goal(goals) ) 
		return result;
	else
		return Plan::create_infinity_plan();
} */


Plan GoalSearch::solve_one_accompany_goal(
							State& state, SearchContext& context, unsigned depth){
	unsigned fact,sfact, i,j;
	unsigned flag = true;
	Plan result;
	result.set_empty();
	result.es.clone( state );
	
	vector<unsigned> ffs;
	for(j=0;j<context.num_fixed_fact;j++) ffs.push_back( context.fixed_fact[j]);
	
	for(i=0;i<goals.size();i++){
		if( state.gvalue[task.gft[goals[i]].graph] == task.gft[goals[i]].vertex)
			continue;
		if( !task.test_reachability( state, ffs, goals[i] ) ) 
			return Plan::create_infinity_plan(); 
	}
	
	for(i=0;i<goals.size();i++){
		vector<unsigned> slave_goals;
		fact = goals[i];
		if( state.gvalue[task.gft[fact].graph] == task.gft[fact].vertex)
			continue;

		if( task.solverContext->debug == true ){
			cout<<depth<<",Solving sub goal:";
			task.print_ft_name(fact);
			cout<<endl;
		}
		
		for(j=0;j<task.aco_pr_vec.size();j++)
			if( fact == task.aco_pr_vec[j].first ) break;
		
		//This change is to handle the case that the ???.second vec is empty;
		if( j < task.aco_pr_vec.size() && task.aco_pr_vec[j].second.size()!= 0 )
			slave_goals = task.aco_pr_vec[j].second;
		else
			slave_goals.push_back(fact);
		flag = false;
		
		for(j=0;j<slave_goals.size();j++){
			sfact = slave_goals[j];
			
			TransitionGraph& graph = task.transGraph[task.gft[sfact].graph];
			DTGSearch dtgsearch( graph, context );
			dtgsearch.set_dest( sfact );
			context.cg_index.clear();
			result.set_empty();
			result.es.clone( state );			
			Plan soln = dtgsearch.solve( result );	
			
			if( soln.is_infinity() ) continue;
			
			context.add_ff( sfact );
			Plan other = solve_one_accompany_goal( soln.es, context, depth + 1);
			context.num_fixed_fact--;
			
			if( !other.is_infinity() && other.es.check_goal(goals) ) {
				result.set_empty();
				result.merge( soln );
				result.merge( other );
				result.es = other.es;
				
				if( depth == 0 )
					update_solution( result );
				else
					return result;
			}
		}
	}
	
	//Solve all the other goals; It is simply incremental and not complete;
	if( flag ){
		for(i=0;i<goals.size();i++){
			fact = goals[i];
			if( state.gvalue[task.gft[fact].graph] == task.gft[fact].vertex)
				continue;
			for(j=0;j<task.aco_pr_vec.size();j++)
				if( fact == task.aco_pr_vec[j].first ) break;
			if( j < task.aco_pr_vec.size()) continue;
			
			TransitionGraph& graph = task.transGraph[task.gft[fact].graph];
			DTGSearch dtgsearch( graph, context );
			context.cg_index.clear();
			dtgsearch.set_dest( fact );
			result = dtgsearch.solve( result );
		}
		
		//Only the deepest recursive function could return with valid value;
		if( result.es.check_goal(goals) ) 		
			return result;
	}
	
	return Plan::create_infinity_plan();
}


///Strategy#2. Atomic sequential solving for accompany goals;
// This routine is supposed to be able to be combined with the first approach, no time in this version;
Plan GoalSearch::solve_atomic_accompany_goals(State& state){
	
	CloseList close_list;
	SearchContext context( close_list );
	context.do_permutation = false;	
	Fringe fringe( context, this->goals );		//This fringe is dummy
	context.fringe_ptr = &fringe;
	
	cout<<"Goal Solving strategy: Atomic Incremental For AccoGoals"<<endl;
	
	solve_one_accompany_goal(state, context, 0 );
	
	return this->best_soln;
}


/// Solving strategies,  #1, Brute Force;
/// Just for debug use in this version;
/// This method reads goal agenda info directly from task.goals_agenda;
Plan GoalSearch::solve_atomic_fixed_order(State& state){

	CloseList close_list;
	SearchContext context(close_list);
	context.do_permutation = true;
	Fringe fringe( context, this->goals );
	context.fringe_ptr = &fringe;
		
	cout<<"Goal Solving strategy: Atomic_Fixed_Order ...";		
	if( context.debug ) printf("\n***CG_Search Debug_Start***\n"); 
	
	Plan result;
	result.set_empty();
	result.es.clone(state);
	
	unsigned i,j,fact = UNDEFINED;
	for( i=0 ; i<task.goals_agenda.size() ;i++)
	for( j=0; j<task.goals_agenda[i].size() ;j++ ){
		fact = task.goals_agenda[i][j];
		if( task.solverContext->debug == true ){
			cout<<"Solving sub goal:";
			task.print_ft_name(fact);
			cout<<endl;
		}
		
		assert( task.gft[fact].graph!=UNDEFINED );		
		TransitionGraph& graph = task.transGraph[task.gft[fact].graph];
		DTGSearch dtgsearch( graph, context );
		dtgsearch.set_dest( fact );
		//cout<<"I "<<i<<" "<<j<<endl;
		//if( i== 0 && j == 0  ) context.print_level = 30;
		Plan  solv = dtgsearch.solve( result );
		//if( i ==0 && j == 0 ) exit(0);
		
		if( solv.is_infinity() ){
			cout<<"  Fail."<<endl;
			return Plan::create_infinity_plan();
		}
		if( task.solverContext->debug == true ){
			task.print_ft_name( fact );
			printf("One subgoal done!!!!\n");
			//context.print_ff_list();
			solv.println("Partial plan:");
			solv.es.println("STATE:");
			cout<<endl;
		}
			
		result.merge( solv );
		result.es = solv.es;
		context.add_ff(fact);
	}
	
	return result;
}


bool GoalSearch::update_solution(Plan& plan){
	
	plan.evaluate_cost();
	if( plan.cost() >= this->best_soln.cost() ) 
		return false;
	
	this->soln_count++;
	char new_file[MAX_FILENAME];
	sprintf( new_file , "%s.%d", task.solverContext->soln_file, soln_count );
	cout<<"Writing an updated(better) solution to file "<<new_file
	     	<<" with cost "<<plan.cost()<<endl;
	
	plan.write_to_file( new_file );
	this->best_soln = plan;
	return true;
}


//Determine the hierachical levels of graphs;
//The level 0 is always those goal-related graphs;
unsigned determine_graph_levels(vector<unsigned>& goals){
	
	unsigned 	i,j,gid;
	bool 		processed[MAX_DTG_NUM];
	unsigned 	dlevel = 0;
	memset( processed, false, sizeof(bool) * task.transGraphNum );
	
	bool flag[MAX_DTG_NUM][MAX_DTG_NUM];
	for( i=0; i<task.transGraphNum; i++)
		for(j=0;j<task.transGraphNum;j++)
			flag[i][j] = task.causalMatrix[i][j];	
	
	for( i=0;i<goals.size();i++){
		gid = task.gft[goals[i]].graph;
		task.transGraph[gid].dlevel = 0;
		
		for(j=0;j<task.transGraphNum;j++)
			flag[gid][j] = flag[j][gid] = false;
		processed[gid] = true;
	}
	
	//Level#1 is also pre-defined, if certain conditions hold;
	//	1. It is depended by a level#0 graph;
	//	2. it is not a leaf; i.e. depends on at least another graph;
	//  3. It is not a NONE- related graph;
	for( i=0;i<task.transGraphNum; i++ ){
		if( task.transGraph[i].dlevel != UNDEFINED )
			continue;
		
		bool cond1 = false;
		for( j=0;j<task.transGraphNum;j++)
			if( task.transGraph[j].dlevel == 0 && task.causalMatrix[j][i] ){
				cond1 = true;
				break;
			}
		if( !cond1 ) continue;
		
		bool cond2 =  false;
		for(j=0;j<task.transGraphNum;j++)
			if( task.causalMatrix[i][j] ){
				cond2 = true;
				break;
			}
		
		if( cond1 && cond2 && !task.transGraph[i].hasNegateAll ){
			dlevel = 1;
			task.transGraph[i].dlevel= dlevel;
			for(j=0;j<task.transGraphNum;j++)
				flag[i][j] = flag[j][i] = false;
			processed[i] = true;
		}
	}
	
	do{
		dlevel++;
		vector<unsigned> picked_list;
		for(i=0;i<task.transGraphNum;i++){
			if( processed[i] ) continue;
			for( j=0; j<task.transGraphNum; j++ )
				if( flag[j][i] ) break;
			
			if( j == task.transGraphNum ){
				//for(k=0;k<task.transGraphNum;k++)
				//	if( flag[i][k] ) break;
				//if( k< task.transGraphNum )
				picked_list.push_back(i);
			}
		}
		/*
		cout<<"PICKED LIST"<<endl;
		for(i=0;i<picked_list.size();i++)
			cout<<picked_list[i]<<";";
		cout<<endl;
		*/
		if( picked_list.size() == 0 ){//If no 
			int max = -10000;
			for( i = 0 ; i < task.transGraphNum ; i++ ){
				if( processed[i]) continue;

				TransitionGraph& graph = task.transGraph[i]; 
				int v = graph.dependToDegree - graph.dependByDegree;
				if( v > max ) max = v;
			}
			
			for( i = 0 ; i < task.transGraphNum ; i++ ){
				if( processed[i]) continue;
				
				TransitionGraph& graph = task.transGraph[i]; 
				int v = graph.dependToDegree - graph.dependByDegree;
				if( v == max ) picked_list.push_back( i );
			}
		}
		
		if( picked_list.size() == 0 ){
			dlevel--;		//cancel the last scaling up;
			break;
		}
		
		for(i=0;i<picked_list.size();i++){
			unsigned gid = picked_list[i]; 
			task.transGraph[gid].dlevel = dlevel;
			
			for(j=0;j<task.transGraphNum;j++)
				flag[gid][j] = flag[j][gid] = false;
			processed[gid] = true;		
		}
	}while( true );	//will break out above;

	
	if( task.solverContext->display_info == 200 ){
		for(i=0;i<task.transGraphNum;i++){
			if( task.transGraph[i].dlevel == UNDEFINED ){
				cout<<"ERR: Before search, one graph not assigned with dlevel"<<endl;
				exit(0);
			}
			task.transGraph[i].print_vertex_only();
		}
		exit(0);
	}
	return dlevel;
}

//End of file


