/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include<vector>
#include<cassert>

#include "state.h"
#include "plan.h"
#include "task.h"
#include "utils.h"

using namespace std;


static Task& task = Task::getInstance();

Plan::Plan(){
	infinity = false;
	total_cost = 0;
}


Plan::Plan(const Plan& p){
	infinity = p.infinity;
	seq = p.seq;
	total_cost = p.total_cost;
	es = p.es;
}


Plan::~Plan(){
	//Nothing else to do;
	seq.clear();
}

/*  deep copy */
void Plan::clone(const Plan& p){

	if( p.infinity )
		this->set_infinity();
	else
		for(unsigned i=0;i<p.seq.size();i++)
			this->seq.push_back( p.seq[i] );

	es.clone( p.es );
	total_cost = p.total_cost;
}

void Plan::copy(Plan& from){
	infinity = from.infinity;
	seq.clear();
	for(unsigned i=0;i<from.length();i++)
		seq.push_back( from.seq[i] );
	total_cost = from.total_cost;
}


Plan& Plan::operator=(const Plan& p){
	if( this == &p ) return *this;
	
	if( p.infinity )
		infinity = true;
	else{
		infinity = p.infinity;
		seq = p.seq;
		es = p.es;
		this->total_cost = p.total_cost;
	}
	
	return *this;
}

unsigned& Plan::operator[](unsigned offset){
	return seq[offset];
}




bool Plan::merge(Plan& to_be_merged){
	if( to_be_merged.infinity )
		infinity = true;
	else{
		for(unsigned i=0;i<to_be_merged.length();i++)
			seq.push_back( to_be_merged.seq[i] );
		total_cost += to_be_merged.cost();
	}
		return true;
}


void Plan::add_oper( unsigned op ){
	assert( !infinity );
	seq.push_back( op );
	total_cost += task.gop[op].cost;
}


void Plan::set_infinity(){
	seq.clear();
	infinity = true;
	total_cost = INF;
}


void Plan::set_empty(){
	infinity = false;
	seq.clear();
	total_cost = 0;
}

unsigned Plan::cost(){
	return this->total_cost;
}

unsigned Plan::evaluate_cost(){
	this->total_cost = 0;
	for(unsigned i=0;i<seq.size();i++)
		this->total_cost += task.gop[seq[i]].cost;
	return this->total_cost;
}


/*----------------------------------------------------------------------------------*/



bool Plan::is_empty(){
	return ( !infinity && length() ==0 );
}

bool Plan::is_infinity(){
	return infinity;	
}

Plan& Plan::create_infinity_plan(){
	static Plan inf_plan;
	inf_plan.infinity = true;
	return inf_plan;
}

Plan& Plan::create_empty_plan(){
	static Plan ept_plan;
	assert( ept_plan.length() == 0 );
	ept_plan.infinity = false;
	ept_plan.seq.clear();	
	return ept_plan;
}


unsigned Plan::length(){
	if( is_infinity() ) return INF;

	return seq.size();
}


void Plan::print(){
	Task& task = Task::getInstance();
	cout<<"[PL:"<<length()<<"]";
	unsigned i;
	for(i=0;i<seq.size();i++){
		task.print_op_name( seq[i] );
		cout<<";";
	}
}


void Plan::println(){
	print();
	cout<<endl;
}


void Plan::println(const char* str){
	cout<<str;
	print();
	cout<<endl;
}

void Plan::write_to_file(const char* filename){
	
	FILE* solnFile;
	if ( (solnFile = fopen(filename, "w") ) == NULL ) {
        printf("\n\nEXIT: can't open file to write final solution.\n\n");
        exit(1);
    }
    
    fprintf( solnFile, "; Time %.3f\n",  get_global_time());	
    fprintf( solnFile, "; nrActions %u\n\n", length() );
    
	unsigned i, action = UNDEFINED;
	for( i=0; i<length() ; i++ ){
        fprintf( solnFile, "%d: ", i);
        action = seq[i];
        
        assert( action>=0 && action< task.gnum_op );
		fprintf( solnFile, task.gop[action].name.c_str() );
        fprintf( solnFile, " [1]\n"); 
	}
	
	fclose( solnFile );
}


