/**********************************
 * R. Huang, Y. Chen and W. Zhang
 * Washington University in STL
 * MAY/2008
 *********************************/

#include<iostream>
#include "dp.h"
#include "state.h"
#include "plan.h"
#include "task.h"
#include "transition_graph.h"
#include "solver_context.h"
#include "resource_allocator.h"
#include "utils.h"

using namespace std;


static Task&task = Task::getInstance();
static ResourceAllocator& ra = ResourceAllocator::getInstance();
	
State::State(){

	gvalue = ra.allocate_state(pool_pos);
	ra.inc_state_counter( pool_pos );
}

State::State(const State& s){
	gvalue = s.gvalue;
	pool_pos = s.pool_pos;
		
	ra.inc_state_counter( pool_pos );
}

State::State(vector<unsigned>& init){
	
	gvalue = ra.allocate_state(pool_pos);
	ra.inc_state_counter( pool_pos );
		
	translate_from_strips( init );
}

State& State::operator=(const State& p){
	if( this == &p ) return *this;
		
	ra.dec_state_counter( pool_pos );
	
	gvalue = p.gvalue;
	pool_pos = p.pool_pos;
	ra.inc_state_counter( pool_pos );
		
	return *this;
} 

State::~State(){
	ra.dec_state_counter( pool_pos );
}

void State::clone(const State& p){
	if( gvalue == NULL || p.gvalue == 0 ){
		cout<<"ERR:Please call alloc() before any assignment."<<endl;
		exit(0);
	}
	for(unsigned i=0;i<task.transGraphNum;i++)
		gvalue[i] = p.gvalue[i];
} 



unsigned& State::operator[](unsigned offset){
	return this->gvalue[offset];	
}


bool State::check_goal(vector<unsigned>& goals){
	unsigned i,fid,gid;
	
	for(i=0;i<goals.size();i++){
		
		fid = goals[i];
		gid = task.gft[fid].graph;
		
		if (  gvalue[gid] != task.gft[fid].vertex ){
			//task.print_ft_name( fid );cout<<" is not satisfied"<<endl;
			return false;
		}
	}
	
	return true;
}



/*------------------------------------------------------------------------------------*/


void State::translate_from_strips(vector<unsigned>& factset){
	assert( gvalue!= NULL );
	unsigned i,j;
	memset( this->gvalue , UNDEFINED , sizeof(unsigned)*task.transGraphNum );
	for( i=0; i<factset.size(); i++ ){
		
		unsigned graphId = task.gft[factset[i]].graph;
		/* Some of the facts might not in the SAS+ formulation, because FW considers 
		 * them as unnecessary and remove them */
		if( graphId == UNDEFINED ) continue;
		
		gvalue[ graphId ] = task.gft[factset[i]].vertex;
	}
	
	//Handle remaining graphs 
	for( i=0;i<task.transGraphNum;i++){
		if( gvalue[i] != UNDEFINED ) continue;
		TransitionGraph& graph = task.transGraph[i];
		if( !graph.isAxiom ){
			//For normal graph, assign it with NONE- fact;
			for( unsigned j=0; j<graph.size ; j++){
				if( graph.vertex[j].isNone() ){
					gvalue[i] = j;
					break;
				}
			}
		}else{
			//Hack for axioms
			assert( graph.size == 2 );
			assert( graph.vertex[1].isNone() );
			gvalue[i] = 0;
		}
	}
	
	//Apply rules immediately;
	for(i=task.gnum_op;i<task.gpos_xm;i++){
		Operator& oper = task.gop[i];
		bool xm_sat = true;
		for(j=0;j<oper.num_T;j++){
			//task.transGraph[oper.Tg[j]].print_vertex(oper.Tp[j]);
			if( oper.Tp[j] == NO_PRE_NEEDED ) continue;
			if( gvalue[oper.Tg[j]] != oper.Tp[j]){
				xm_sat = false;
				break;
			}
		}

		if( xm_sat ){
			for(j=0;j<oper.num_T;j++){
				gvalue[oper.Tg[j]] = oper.Te[j];
			}
		}
		/*else{
			assert( false );			//Supposed to be not reachable here;
		} */
	}
	
	if( task.solverContext->debug ){													//Debug
		for(i=0;i<task.transGraphNum;i++){
			unsigned index = gvalue[i];
			if( index > task.transGraph[i].size ){
				cout<<"Error in translating state."<<endl;
				task.transGraph[i].print();
				exit(1);
			}
		}
		cout<<endl;

		for( i=0;i<task.transGraphNum;i++){
			if( !task.transGraph[i].isAxiom ) continue;
			task.transGraph[i].print_vertex(gvalue[i]);
			if( !task.transGraph[i].vertex[gvalue[i]].isNone() ){
				cout<<"There is non-static axiom in this domain. To-DO:"<<endl;
				exit(0);
			}
		}
	}																													//Debug
	
}


bool State::apply_action_effect(unsigned operId){
	
	Operator& oper = task.gop[operId];
	assert( gvalue != NULL );
	
	unsigned i,j,g;
	
	for(i=0;i<oper.num_T;i++){
		g = oper.Tg[i];
		
		if( oper.Tp[i] != NO_PRE_NEEDED ){
			if( gvalue[g] != oper.Tp[i]){
				printf("\n Critical Err: While applying Action[%d]:",operId);
				oper.print_name();printf("\n");
				printf("The precondition in graph %d is supposed to be [v%d]", g, oper.Tp[i]);
				task.transGraph[g].vertex[oper.Tp[i]].print();
				printf("But is now [v%d] ",gvalue[g]);
				task.transGraph[g].vertex[gvalue[g]].print();
				cout<<endl;
				task.gop[operId].print_detail();
				fflush(stdout);
				task.gop[operId].print_as_sas();
				exit(0);
			}
			assert ( gvalue[g] == oper.Tp[i] );
		}
		gvalue[g] = oper.Te[i];
	}
	
	unsigned fact = UNDEFINED;
	//Apply rules immediately;
	for(i=task.gnum_op;i<task.gpos_xm;i++){
		Operator& oper = task.gop[i];
		bool xm_sat = true;
		for(j=0;j<oper.num_P;j++){
			fact = oper.P[j];
			if( gvalue[task.gft[fact].graph] != task.gft[fact].vertex){
				xm_sat = false;
				break;
			}
		}
		assert( oper.num_A == 1);		
		fact = oper.A[0];

		if( xm_sat ){
			gvalue[task.gft[fact].graph] = task.gft[fact].vertex;
			assert( task.transGraph[task.gft[fact].graph].
						vertex[task.gft[fact].vertex].isNone() );
		}else{
			unsigned not_triggered = 
				task.transGraph[task.gft[fact].graph].vertex[0].isNone() ? 1 : 0;
			gvalue[task.gft[fact].graph] = not_triggered;
		}
	}
	
	return true;
}



unsigned State::hash_value(){

	/*
	unsigned hash = 0;
	
	for(unsigned i=0;i<task.transGraphNum;i++){
		unsigned fact = task.transGraph[i].vertex[gvalue[i]][0];
		hash += task.gft[fact].rand;
	}
	
	return hash; */
	
	
	return DJBHash( gvalue, task.transGraphNum );
}

void State::print(){

	if( gvalue == NULL ) cout<<"INVALID State."<<endl;
	printf("STATE:");
	for( unsigned i=0; i<task.transGraphNum ; i++ ){
		unsigned v = this->gvalue[i];
		printf("[%d:%u]",i, v);
		task.transGraph[i].print_vertex(v);
	}
}


void State::println(const char* str){
	cout<<str;
	print();
	printf("\n");
}


bool State::is_valid(){
	for(unsigned i=0;i<task.transGraphNum;i++){
		if( gvalue[i] == UNDEFINED ) return false;
		if( gvalue[i] <0 || gvalue[i]>=task.transGraph[i].size )	
			return false;
	}
	return true;
}


bool State::isSatisfied(unsigned* array, unsigned size){
	
	for(unsigned i=0;i<size;i++){
		unsigned fid = array[i];
		unsigned graphId = task.gft[fid].graph;
		if( gvalue[graphId] != task.gft[fid].vertex ) 
			return false;
	}
	
	return true;
}

//END of file;

