#! /usr/bin/env python
# -*- coding: latin-1 -*-

import build_model
import pddl_to_prolog
import pddl

def get_fluent_facts(task, model):
  cost_func_names = []
  for pred in task.cost_predicates:
    cost_func_names.append( pred.name )

  fluent_predicates = set()
  for action in task.actions:
    for effect in action.effects:
      if not effect.literal.predicate in cost_func_names:    
        fluent_predicates.add(effect.literal.predicate)

  for axiom in task.axioms:
    fluent_predicates.add(axiom.name)
  return set([fact for fact in model
              if fact.predicate in fluent_predicates])

def get_fluent_functions(task, model):
  #for atom in model:
  #  print atom
  cost_func_names = []
  for pred in task.cost_predicates:
    cost_func_names.append( pred.name )
  
  fluent_predicates = set()
  for action in task.actions:
    for effect in action.effects:
      if effect.literal.predicate in cost_func_names:
        fluent_predicates.add(effect.literal.predicate)
  return [fact for fact in model
              if fact.predicate in fluent_predicates]

def get_objects_by_type(typed_objects):
  result = {}
  for obj in typed_objects:
    result.setdefault(obj.type, []).append(obj.name)
  return result

def instantiate(task, model):
  relaxed_reachable = False
  fluent_facts = get_fluent_facts(task, model)
  #R.Huang
  cost_functions = get_fluent_functions( task, model )
  cost_pred_name = [pred.name for pred in task.cost_predicates]
  #Update initial state, remove function values and update to cost_functions
  to_del_list = []
  for i,fact in enumerate(task.init):
    if fact.predicate in cost_pred_name:
      #print fact, fact.value
      to_del_list.append(i)
      for cost_func in cost_functions:
        if cost_func == fact:
          cost_func.value = fact.value

  to_del_list.reverse()
  for pos in to_del_list:
    del task.init[pos]
  init_facts = set(task.init)

  type_to_objects = get_objects_by_type(task.objects)

  instantiated_actions = []
  instantiated_axioms = []
  for atom in model:
    if isinstance(atom.predicate, pddl.Action):
      action = atom.predicate
      parameters = action.parameters
      if isinstance(action.precondition, pddl.ExistentialCondition):
        parameters = list(parameters)
        parameters += action.precondition.parameters
      variable_mapping = dict([(par.name, arg)
                               for par, arg in zip(parameters, atom.args)])
      inst_action = action.instantiate(variable_mapping, init_facts,
                                       fluent_facts, type_to_objects, cost_functions)
      if inst_action:
        instantiated_actions.append(inst_action)
    elif isinstance(atom.predicate, pddl.Axiom):
      axiom = atom.predicate
      parameters = axiom.parameters
      if isinstance(axiom.condition, pddl.ExistentialCondition):
        parameters = list(parameters)
        parameters += axiom.condition.parameters
      variable_mapping = dict([(par.name, arg)
                               for par, arg in zip(parameters, atom.args)])
      inst_axiom = axiom.instantiate(variable_mapping, init_facts, fluent_facts)
      if inst_axiom:
        instantiated_axioms.append(inst_axiom)
    elif atom.predicate == "@goal-reachable":
      relaxed_reachable = True
  
  return relaxed_reachable, fluent_facts, instantiated_actions, instantiated_axioms

def explore(task):
  prog = pddl_to_prolog.translate(task)
  model = build_model.compute_model(prog)
  return instantiate(task, model)

if __name__ == "__main__":
  import pddl

  task = pddl.open()
  relaxed_reachable, atoms, actions, axioms = explore(task)
  print "goal relaxed reachable: %s" % relaxed_reachable
  print "%d atoms:" % len(atoms)
  for atom in atoms:
    print " ", atom
  print
  print "%d actions:" % len(actions)
  for action in actions:
    action.dump()
    print
  print
  print "%d axioms:" % len(axioms)
  for axiom in axioms:
    axiom.dump()
    print
