import actions
import axioms
import conditions
import predicates
import pddl_types
import string

class Task(object):
  def __init__(self, domain_name, task_name, requirements,
               types, objects, predicates, init, goal, actions, axioms, cost_predicates):
    self.domain_name = domain_name
    self.task_name = task_name
    self.requirements = requirements
    self.types = types
    self.objects = objects
    self.predicates = predicates
    self.init = init
    self.goal = goal
    self.actions = actions
    self.axioms = axioms
    self.axiom_counter = 0
    #R.Huang
    self.cost_predicates = cost_predicates
    self.cost_enabled = False
    for requirement in requirements.requirements:
      if string.find(requirement,":action-costs") >= 0:
        self.cost_enabled = True
        break;

  def add_axiom(self, parameters, condition):
    name = "new-axiom@%d" % self.axiom_counter
    self.axiom_counter += 1
    axiom = axioms.Axiom(name, parameters, condition)
    self.predicates.append(predicates.Predicate(name, parameters))
    self.axioms.append(axiom)
    return axiom

  def parse(domain_pddl, task_pddl):
    domain_name, requirements, types, constants, predicates, functions, actions, axioms \
                 = parse_domain(domain_pddl)

    task_name, task_domain_name, objects, init, goal = parse_task(task_pddl,requirements)
    
    assert domain_name == task_domain_name
    objects = constants + objects
    init += [conditions.Atom("=", (obj.name, obj.name)) for obj in objects]
    return Task(domain_name, task_name, requirements, types, objects,
                predicates, init, goal, actions, axioms, functions)
  parse = staticmethod(parse)

  def dump(self):
    print "Problem %s: %s [%s]" % (self.domain_name, self.task_name,
                                   self.requirements)
    print "Types:"
    for type in self.types:
      print "  %s" % type
    print "Objects:"
    for obj in self.objects:
      print "  %s" % obj
    print "Predicates:"
    for pred in self.predicates:
      print "  %s" % pred
    print "Init:"
    for fact in self.init:
      print "  %s" % fact
    print "Goal:"
    self.goal.dump()
    print "Actions:"
    for action in self.actions:
      action.dump()
    if self.axioms:
      print "Axioms:"
      for axiom in self.axioms:
        axiom.dump()

class Requirements(object):
  def __init__(self, requirements):
    self.requirements = requirements
    for req in requirements:
      assert req in (
        ":strips", ":adl", ":typing", ":negation", ":equality",
        ":negative-preconditions", ":disjunctive-preconditions",
        ":quantified-preconditions", ":conditional-effects", ":derived-predicates",":action-costs"), req
  def __str__(self):
    return ", ".join(self.requirements)

def parse_domain(domain_pddl):
  iterator = iter(domain_pddl)

  assert iterator.next() == "define"
  domain_line = iterator.next()
  assert domain_line[0] == "domain" and len(domain_line) == 2
  yield domain_line[1]

  opt_requirements = iterator.next()
  action_cost_enabled = False
  if opt_requirements[0] == ":requirements":
    requirement = Requirements(opt_requirements[1:])
    if string.find(requirement.__str__(),"action-costs")>=0:
      action_cost_enabled = True
    yield requirement
    opt_types = iterator.next()
  else:
    yield Requirements([":strips"])
    opt_types = opt_requirements

  the_types = [pddl_types.Type("object")]
  if opt_types[0] == ":types":
    the_types.extend(pddl_types.parse_typed_list(opt_types[1:],
                                                 constructor=pddl_types.Type))
    opt_constants = iterator.next()
  else:
    opt_constants = opt_types
  pddl_types.set_supertypes(the_types)
  # for type in the_types:
  #   print repr(type), type.supertype_names
  yield the_types

  if opt_constants[0] == ":constants":
    yield pddl_types.parse_typed_list(opt_constants[1:])
    pred = iterator.next()
  else:
    yield []
    pred = opt_constants

  assert pred[0] == ":predicates"
  yield ([predicates.Predicate.parse(entry) for entry in pred[1:]] +
         [predicates.Predicate("=",
                               [pddl_types.TypedObject("?x", "object"),
                                pddl_types.TypedObject("?y", "object")])])
  
  #R.Huang
  if action_cost_enabled:
    functions = iterator.next()
    cost_functions = []
    assert functions[0] == ":functions"
    count = 0;
    for entry in functions[1:]:
      if count % 3 == 0:
        if string.find( entry[0], "total-cost" ) >=0:
          None
        else:
          cost_functions.append( predicates.Predicate.parse(entry))
          #cost_functions.append( entry[0]);
      elif count % 3 == 1:
        assert( string.find( entry,'-')>=0 );
      else:
        assert( string.find( entry,'number')>=0 )
      count = count + 1
    yield cost_functions
  else:
    yield []

  the_axioms = []
  the_actions = []
  for entry in iterator:
    if entry[0] == ":derived":
      axiom = axioms.Axiom.parse(entry)
      the_axioms.append(axiom)
    else:
      action = actions.Action.parse(entry)
      the_actions.append(action)
  yield the_actions
  yield the_axioms


def parse_task(task_pddl,requirement):
  iterator = iter(task_pddl)

  assert iterator.next() == "define"
  problem_line = iterator.next()
  assert problem_line[0] == "problem" and len(problem_line) == 2
  yield problem_line[1]
  domain_line = iterator.next()
  assert domain_line[0] == ":domain" and len(domain_line) == 2
  yield domain_line[1]

  objects_opt = iterator.next()
  if objects_opt[0] == ":objects":
    yield pddl_types.parse_typed_list(objects_opt[1:])
    init = iterator.next()
  else:
    yield []
    init = objects_opt

  assert init[0] == ":init"
  init_facts = []
  for fact in init[1:]:
    if fact[0] == "=":
      real_fact = fact[1];
      func_unit = conditions.Atom(real_fact[0], real_fact[1:])
      func_unit.value = int(fact[2])
      init_facts.append(func_unit)
    else:
      init_facts.append( conditions.Atom(fact[0], fact[1:]) )
  #yield [conditions.Atom(fact[0], fact[1:]) for fact in init[1:]]
  yield init_facts
  
  goal = iterator.next()
  assert goal[0] == ":goal" and len(goal) == 2
  yield conditions.parse_condition(goal[1])
  
  if string.find( requirement.__str__(), "action-costs")>=0:
    metric = iterator.next()
    #R.Huang Assume the metric is always minimizing the total-cost;
    assert metric[0] == ":metric" and len(metric) == 3
    assert metric[1] == "minimize"
    assert metric[2][0] == "total-cost"
  
  for entry in iterator:
    assert False, entry
